import httpx
from rdflib import URIRef
from .object import FedoraObject
import os
import base64


class FedoraFile(FedoraObject):
    def __init__(self, uri, position=0):
        super().__init__(uri)
        self.position = position

    def get_files(self):
        return list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.pcdm}hasFile")
        ))

    def get_contains(self):
        return list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.ldp}contains")
        ))

    def get_cantaloupe_base_64(self):
        server = self.uri.split('/')[2]
        encoded = base64.urlsafe_b64encode(bytes(self.uri, "utf-8"))
        decoded = encoded.decode("utf-8")
        return f"https://{server}/iiif/2/{decoded}/info.json"

    def download(self, path):
        file_path = str(list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.pcdm}hasFile")
        ))[0])
        if not os.path.exists(path):
            os.makedirs(path)
        with httpx.Client() as client:
            response = client.get(str(file_path))
            response.raise_for_status()
            with open(f"{path}/{self.position.zfill(4)}_{file_path.split('/')[-1]}", 'wb') as file:
                for chunk in response.iter_bytes():
                    file.write(chunk)

    def get_parts(self):
        return list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.dcterms}hasPart")
        ))


import click
from . import FedoraWork
from .fedora import FedoraCollection
from csv import DictWriter
from tqdm import tqdm


@click.group()
def cli():
    pass


@cli.command()
@click.argument("uri")
def ferry(uri):
    collection = FedoraCollection(uri)
    members = collection.get_members()

    rows = []
    for member in tqdm(members):
        work = FedoraWork(member)
        row = {k: "|".join(v) for k, v in work.metadata_to_dict().items()}
        row["files"] = "|".join(work.get_ordered_member_files())
        rows.append(row)

    fieldnames = list(dict.fromkeys(k for row in rows for k in row))
    with open(f"{collection.uri.split('/')[-1]}.csv", "w", newline="", encoding="utf-8") as f:
        writer = DictWriter(f, fieldnames=fieldnames, extrasaction="ignore")
        writer.writeheader()
        writer.writerows(rows)


@cli.command()
@click.argument("uri")
def get_members(uri):
    all_collections = []
    collection = FedoraCollection(uri)
    for member in tqdm(collection.get_contains()):
        if "_objects" not in member:
            current = FedoraCollection(member)
            all_collections.append(
                {
                    "uri": str(member),
                    "created": str(current.get_created()),
                    "works": len(current.get_members()),
                }
            )
    with open("collections.csv", "w", newline="", encoding="utf-8") as f:
        writer = DictWriter(f, fieldnames=list(all_collections[0].keys()))
        writer.writeheader()
        writer.writerows(all_collections)


if __name__ == "__main__":
    cli()



from .object import FedoraObject
from rdflib import URIRef


class FedoraProxy(FedoraObject):
    def __init__(self, uri):
        super().__init__(uri)
        self.proxy_for = list(self.content.objects(
            subject=URIRef(uri), predicate=URIRef(f"{self.namespaces.ore}proxyFor")
        ))
        self.iana_next = list(self.content.objects(
            subject=URIRef(uri), predicate=URIRef(f"{self.namespaces.iana}next")
        ))
        self.has_proxy_for = self.check_exists(self.proxy_for)
        self.has_iana_next = self.check_exists(self.iana_next)

    @staticmethod
    def check_exists(value):
        return len(value) > 0

    def get_next(self):
        return str(self.iana_next[0]) if self.iana_next else None

    def get_proxy_for(self):
        return str(self.proxy_for[0]) if self.proxy_for else None


from .object import FedoraObject
from .proxy import FedoraProxy
from .file import FedoraFile
from rdflib import URIRef, Graph


class FedoraWork(FedoraObject):
    def __init__(self, uri):
        super().__init__(uri)
        self.uri = URIRef(uri)
        self.descriptive_metadata = self.__get_descriptive_metadata()
        self._cache = {}
        self.iana_first = self.__get_iana('first')
        self.iana_last = self.__get_iana('last')

    def __get_iana(self, name):
        return next(self.content.objects(subject=self.uri, predicate=URIRef(f"{self.namespaces.iana}{name}")), None)

    def __get_descriptive_metadata(self):
        """Builds a graph with just known descriptive metadata."""
        description = Graph()
        descriptive_predicates = {
            f"{self.namespaces.dc}subject",
            f"{self.namespaces.dc}creator",
            f"{self.namespaces.dc}description",
            f"{self.namespaces.dc}format",
            f"{self.namespaces.dc}language",
            f"{self.namespaces.dc}publisher",
            f"{self.namespaces.dc}title",
            f"{self.namespaces.dc}type",
            f"{self.namespaces.dc}rights",
            f"{self.namespaces.dcterms}alternative",
            f"{self.namespaces.dcterms}created",
            f"{self.namespaces.dcterms}extent",
            f"{self.namespaces.dcterms}isPartOf",
            f"{self.namespaces.dcterms}medium",
            f"{self.namespaces.dcterms}spatial",
        }
        for s, p, o in self.content:
            if str(p) in descriptive_predicates:
                description.add((s, p, o))
        return description

    def __get_cached(self, key, func):
        """Helper to cache expensive operations."""
        if key not in self._cache:
            self._cache[key] = func()
        return self._cache[key]

    @property
    def rights(self):
        return self.__get_cached('rights', lambda: self.__get_first_literal(f"{self.namespaces.dc}rights"))

    @property
    def title(self):
        return self.__get_cached('title', lambda: self.__get_all_literals(f"{self.namespaces.dc}title"))

    @property
    def summary(self):
        return self.__get_cached('summary', lambda: self.__get_first_literal(f"{self.namespaces.dc}description"))

    @property
    def members(self):
        return self.__get_cached('members', lambda: self.__get_all_objects(f"{self.namespaces.pcdm}hasMember"))

    def __get_first_literal(self, predicate):
        for _, p, o in self.content:
            if str(p) == predicate:
                return str(o)
        return ""

    def __get_all_literals(self, predicate):
        return [str(o) for _, p, o in self.content if str(p) == predicate]

    def __get_all_objects(self, predicate):
        return [str(o) for _, p, o in self.content if str(p) == predicate]

    def get_ordered_members(self):
        ordered_members = []
        current = FedoraProxy(uri=str(self.iana_first))
        while current:
            ordered_members.append(current.get_proxy_for())
            if current.has_iana_next:
                current = FedoraProxy(uri=str(current.get_next()))
            else:
                current = None
        return ordered_members

    def get_ordered_member_files(self):
        ordered_members = []
        current = FedoraProxy(uri=str(self.iana_first))
        while current:
            current_part = FedoraFile(uri=str(current.get_proxy_for()))
            current_part_files = current_part.get_files()
            for current_file in current_part_files:
                ordered_members.append(str(current_file))
            if current.has_iana_next:
                current = FedoraProxy(uri=str(current.get_next()))
            else:
                current = None
        return ordered_members

    def get_members(self):
        return list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.pcdm}hasMember")
        ))

    def metadata_to_dict(self):
        description = {}
        for _, p, o in self.content:
            p_str = str(p)
            if p_str.startswith(self.namespaces.dc):
                key = f"dc_{p_str[len(self.namespaces.dc):]}"
            elif p_str.startswith(self.namespaces.dcterms):
                key = f"dcterms_{p_str[len(self.namespaces.dcterms):]}"
            elif p_str.startswith(self.namespaces.fedora):
                key = f"fedora_{p_str[len(self.namespaces.fedora):]}"
            elif p_str.startswith(self.namespaces.ebucore):
                key = f"ebucore_{p_str[len(self.namespaces.ebucore):]}"
            elif p_str.startswith(self.namespaces.ldp):
                key = f"ldp_{p_str[len(self.namespaces.ldp):]}"
            elif p_str.startswith(self.namespaces.premis):
                key = f"premis_{p_str[len(self.namespaces.premis):]}"
            elif p_str.startswith(self.namespaces.relations):
                key = f"relations_{p_str[len(self.namespaces.relations):]}"
            elif p_str.startswith(self.namespaces.tamudigital):
                key = f"tamudigital_{p_str[len(self.namespaces.tamudigital):]}"
            elif p_str.startswith(self.namespaces.tamulocal):
                key = f"tamulocal_{p_str[len(self.namespaces.tamulocal):]}"
            elif p_str.startswith(self.namespaces.title):
                key = f"title_{p_str[len(self.namespaces.title):]}"
            elif p_str.startswith(self.namespaces.schemaorg):
                key = f"schema_{p_str[len(self.namespaces.schemaorg):]}"
            elif p_str.startswith(self.namespaces.rights):
                key = f"rights_{p_str[len(self.namespaces.rights):]}"
            elif p_str.startswith(self.namespaces.subject):
                key = f"subject_{p_str[len(self.namespaces.subject):]}"
            elif p_str.startswith(self.namespaces.identifier):
                key = f"identifier_{p_str[len(self.namespaces.identifier):]}"
            elif p_str.startswith(self.namespaces.identifiertwo):
                key = f"identifiertwo_{p_str[len(self.namespaces.identifiertwo):]}"
            elif p_str.startswith(self.namespaces.foaf):
                key = f"foaf_{p_str[len(self.namespaces.foaf):]}"
            elif p_str.startswith(self.namespaces.bibframe):
                key = f"bibframe_{p_str[len(self.namespaces.bibframe):]}"
            elif p_str.startswith(self.namespaces.oretwo):
                key = f"oretwo_{p_str[len(self.namespaces.oretwo):]}"
            else:
                continue
            description.setdefault(key, []).append(str(o))
        return description


if __name__ == '__main__':
    x = FedoraWork(
        'https://api.library.tamu.edu/fcrepo/rest/3b/6f/c3/25/3b6fc325-f6ca-41d8-b91e-8c5db3be8c13/london-collection_objects/9'
    )




from .object import FedoraObject
from rdflib import URIRef, Graph


class FedoraCollection(FedoraObject):
    def __init__(self, uri):
        super().__init__(uri)
        self.uri = URIRef(uri)
        self._cache = {}

    def get_members(self):
        return list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.pcdm}hasMember")
        ))

    def get_contains(self):
        return list(self.content.objects(
            subject=URIRef(self.uri), predicate=URIRef(f"{self.namespaces.ldp}contains")
        ))

    def get_created(self):
        return next(self.content.objects(subject=self.uri, predicate=URIRef(f"{self.namespaces.fedora}created")), None)


if __name__ == '__main__':
    x = FedoraCollection(
        'https://api.library.tamu.edu/fcrepo/rest/a6/f0/55/01/a6f05501-9312-4fd5-a167-f7ce2fa28ca6'
    )
    print(x.get_members())


import httpx
from rdflib import Graph, Namespace


class Namespaces:
    def __init__(self):
        self.dc = "http://purl.org/dc/elements/1.1/"
        self.pcdm = "http://pcdm.org/models#"
        self.iana = "http://www.iana.org/assignments/relation/"
        self.dcterms = "http://purl.org/dc/terms/"
        self.fedora = "http://fedora.info/definitions/v4/repository#"
        self.rdfs = "http://www.w3.org/1999/02/22-rdf-syntax-ns"
        self.ldp = "http://www.w3.org/ns/ldp#"
        self.ore = "http://www.openarchives.org/ore/terms#"
        self.ebucore = "http://www.ebu.ch/metadata/ontologies/ebucore/ebucore#"
        self.premis = "http://www.loc.gov/premis/rdf/v1#"
        self.relations = "http://purl.org/dc/elements/1.1/relation:"
        self.tamudigital = "http://digital.library.tamu.edu/schemas/"
        self.tamulocal = "http://digital.library.tamu.edu/schemas/local/"
        self.title = "http://purl.org/dc/elements/1.1/title:"
        self.schemaorg = "http://schema.org/"
        self.subject = "http://purl.org/dc/elements/1.1/subject:"
        self.identifier = "http://purl.org/dc/elements/1.1/identifer:"
        self.identifiertwo = "http://purl.org/dc/elements/1.1/identifier:"
        self.rights = "http://purl.org/dc/elements/1.1/rights:"
        self.foaf = "http://xmlns.com/foaf/0.1/"
        self.bibframe = "http://bibframe.org/vocab/"
        self.oretwo = "http://www.openarchives.org/ore/terms/"


class FedoraObject:
    def __init__(self, uri):
        self.uri = uri
        self.namespaces = Namespaces()
        self.content = self.__get_graph()

    def __get_graph(self):
        headers = {
            'Accept': 'application/ld+json'
        }
        r = httpx.get(self.uri, headers=headers, timeout=None)
        try:
            if r.status_code == 200:
                g = Graph()
                g.parse(data=r.content, format='json-ld')
                return g
            else:
                raise Exception(f"Failed to download {self.uri}. Status code: {r.status_code}")
        except:
            print(r.status_code)

    def serialize(self, format="turtle"):
        PCDM = Namespace("http://pcdm.org/models#")
        LDP = Namespace("http://www.w3.org/ns/ldp#")
        IANA = Namespace("http://www.iana.org/assignments/relation/")
        FCREPO = Namespace("http://fedora.info/definitions/v4/repository#")
        ORE = Namespace("http://www.openarchives.org/ore/terms#")
        PREMIS = Namespace("http://www.loc.gov/premis/rdf/v1#")
        EBUCORE = Namespace("http://www.ebu.ch/metadata/ontologies/ebucore/ebucore#")
        DC = Namespace("http://purl.org/dc/elements/1.1/")
        DCTERMS = Namespace("http://purl.org/dc/terms/")
        self.content.bind('pcdm', PCDM)
        self.content.bind('ldp', LDP)
        self.content.bind('iana', IANA)
        self.content.bind('fedora', FCREPO)
        self.content.bind('ore', ORE)
        self.content.bind('premis', PREMIS)
        self.content.bind('ebucore', EBUCORE)
        self.content.bind('dc', DC)
        self.content.bind('dcterms', DCTERMS)
        data = self.content.serialize(format=format)
        return data

    def read_graph(self):
        for s, p, o in self.content:
            print(s, p, o)


from SPARQLWrapper import SPARQLWrapper, JSON
from rdflib import URIRef, Graph, Literal


class FusekiQuery:
    def __init__(self, sparql_endpoint):
        self.sparql_endpoint = SPARQLWrapper(sparql_endpoint)

    def get_predicates_and_objects(self, subject=None, return_graph=False):
        if subject is None:
            s = "?s"
        else:
            s = f"<{subject}>"
        query = f"""
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX ebucore: <http://www.ebu.ch/metadata/ontologies/ebucore/ebucore#>
        SELECT ?p ?o  WHERE {{
            {s} ?p ?o .
        }} 
        """
        self.sparql_endpoint.setQuery(query)
        self.sparql_endpoint.setReturnFormat(JSON)
        results = self.sparql_endpoint.query().convert()
        if return_graph:
            return self.__build_graph(
                subject,
                results["results"]["bindings"]
            )
        return results["results"]["bindings"]

    def get_distinct_last_modified(self):
        query = f"""
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX ebucore: <http://www.ebu.ch/metadata/ontologies/ebucore/ebucore#>
        SELECT DISTINCT ?o  WHERE {{
            ?s <http://fedora.info/definitions/v4/repository#lastModified> ?o .
        }} 
        """
        self.sparql_endpoint.setQuery(query)
        self.sparql_endpoint.setReturnFormat(JSON)
        results = self.sparql_endpoint.query().convert()
        return [value['o']['value'] for value in results["results"]["bindings"]]

    def __build_graph(self, subject, results):
        final = Graph()
        for result in results:
            object_type = result['o']['type']
            if object_type == 'uri':
                object_value = URIRef(result['o']['value'])
            elif object_type == 'literal':
                object_value = Literal(result['o']['value'])
            final.add((URIRef(subject), URIRef(result['p']['value']), object_value))
        return final.serialize(format="turtle", indent=4)


if __name__ == "__main__":
    x = FusekiQuery(sparql_endpoint="https://api-dev.library.tamu.edu/fcrepo-fuseki/fcrepo/query")
    # print(
    #     x.get_predicates_and_objects(
    #         "https://api-dev.library.tamu.edu/fcrepo/rest/3b/6f/c3/25/3b6fc325-f6ca-41d8-b91e-8c5db3be8c13/fb/00/6c/bc/fb006cbc-d2e5-4fcb-9321-62b6385720d0/berger_cloonan_batch_5_objects/241/orderProxies/page_0_proxy",
    #         return_graph=True
    #     )
    # )
    print(x.get_distinct_last_modified())


from .collection import FedoraCollection
from .object import FedoraObject
from .file import FedoraFile
from .work import FedoraWork
from .proxy import FedoraProxy
from .fuseki import FusekiQuery
__all__ = [ "FedoraFile", "FedoraObject", "FedoraCollection", "FedoraProxy", "FedoraWork", "FusekiQuery",]